#
# Licensed Materials - Property of IBM
# Rational ClearQuest 
# (C) Copyright IBM Corp. 2014 All Rights Reserved
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

use CQPerlExt;
use strict;
use warnings;
use Fcntl;

my $action = $ARGV[0];
my @actions = ("help", "import", "export", "validate", "remove");
my $requiredParams = 0;
my @reqiredParamsSet = ("dbset", "userdb", "username", "password");
my %usage = ();

$usage{dbset} = "-";    # dbset
$usage{userdb} = "-";   # userdb name
$usage{username} = "-"; # user
$usage{password} = "-"; # password
$usage{file} = "-";     # file

# See if they requested usage information
if (($#ARGV < 1) || ($ARGV[0] eq "help")) 
{	
	Usage();
}

# Check the number of parameters
if ((($#ARGV != 8) && ($#ARGV != 10)) || ((($action eq "import") || ($action eq "remove")) && ($#ARGV != 10)))
{
	print "\nInvalid number of arguments.\n";
	Usage();
}

# Validate the action in command line
if (!($action~~@actions))
{
	print "\nInvalid command: ".$action."\n";
	Usage();
}

# Get values of the parameters 
for (my $count = 1; $count <= $#ARGV; $count += 2)
{
	my ($d, $switchValue) = split("-", $ARGV[$count]);

	if ($switchValue~~@reqiredParamsSet)
	{
		$requiredParams++;
	}

	if ($usage{$switchValue})
	{
		$usage{$switchValue} = $ARGV[$count + 1];
	}
	else
	{
		print "\nInvalid argument: ".$ARGV[$count]."\n";
		Usage();
	}
}

# Check the required parameters exist
if ($requiredParams ne 4)
{
	print "\nInvalid number of arguments. \n";
	Usage();
}

my $dbSet = $usage{"dbset"};
my $dbName = $usage{"userdb"};
my $user = $usage{"username"};
my $pwd = $usage{"password"};
my $configureFile = $usage{"file"};
my $propertyName = "TYPEAHEAD_PARAMS";
my $siteName = "";	# Typeahead is not site specific
my $adminSessionObj;
my $sessionObj;

ValidateAndLoginUser();

#
# Process the command line
#

if ($action eq "import")
{
	DoCommand_import();
}
elsif ($action eq "export")
{
	DoCommand_export();
}
elsif ($action eq "validate")
{
	DoCommand_validate();
}
elsif ($action eq "remove")
{
	DoCommand_remove();
}
else
{
	# Invalid command line: should never be here but just in case
	print "\nInvalid command: ".$action."\n";
	Usage();
}

print "\n";

# Log off user by closing the sessions
CQAdminSession::Unbuild($adminSessionObj);
CQSession::Unbuild($sessionObj);

exit 0;

# Help information
sub Usage
{
	print <<EOF;

Please see the IBM Rational ClearQuest Information Center for usage details of ClearQuest typeahead application.

EOF
	exit 1;
}

# The typeahead master property looks like this:
#
# users=login_name,fullname,email,misc_info
# Customer=Name,Company,Description,Email
# Project=Name,Description

# The unique key must be the first value of the record type.

# Import and enable the typeahead master property
sub DoCommand_import
{
	my $propertyValue = "";
	my $file = $usage{"file"};
	my @dataFromFile;

	if ((open my $FH, "<", $file))
	{
		@dataFromFile = <$FH>;
		close($FH);

		# Validate the data before import typeahead master property into DB
		# The secord parameter is to indicate whether need to process data before the validation
		my ($result, @data )= ValidateAndCleanTypeaheadData(\@dataFromFile);

		if ($result eq 0)
		{
			die "\nERROR: Validation failed.\n";
		}

		$propertyValue = join("", @data);
		chomp($propertyValue); # Remove the extra blank line at the end of string
		
		eval
		{
			$adminSessionObj->SetMasterProperty($propertyName, $propertyValue, $dbName, $siteName);
			print "\nImported to master property: ".$dbName.".".$propertyName."\n";
		};
		die "\nERROR: $@\n" if $@; $@ = "";
	}
	else
	{
		die "\nERROR: Cannot open file for reading: ".$file."\n".$!."\n";
	}
}

# Export the master property to screen or file
sub DoCommand_export
{
	my $value;

	eval
	{
		$value = $adminSessionObj->GetMasterProperty($propertyName, $dbName, $siteName);
	};
	die "\nERROR: $@\n" if $@; $@ = "";

	if ($usage{"file"} eq "-")
	{
		print "\n---------------------------------------------------------\n";
		print $value;
		print "\n---------------------------------------------------------\n";
	}
	else
	{
		my $file = $usage{"file"};
		if ((open my $FH, ">", $file))
		{
			print $FH $value;
			close $FH;
			print "\nExported to: ".$file."\n";
		}
		else
		{
			die "\nERROR: Cannot open file for writing: ".$file."\n".$!."\n";
		}
	}
}

# Validate the property stored in DB or file
sub DoCommand_validate
{
	my $propertyValue;
	my @dataToValidate;
	
	# Get the Master property
	if ($usage{"file"} eq "-")
	{
		# Get the property value stored in Master Schema
		eval
		{
			$propertyValue = $adminSessionObj->GetMasterProperty($propertyName, $dbName, $siteName);
		};
		die "\nERROR: $@\n" if $@; $@ = "";
		
		@dataToValidate = split(/\r?\n/, $propertyValue);
	}
	else
	{
		# Get the property value stored in the configuration file
		my $file = $usage{"file"};	
		if ((open my $FH, "<", $file))
		{
			@dataToValidate = <$FH>;
			close($FH);
		}
		else
		{
			die "\nERROR: Cannot open file for reading: ".$file."\n".$!."\n";
		}
	}
	my ($result,@data) = ValidateAndCleanTypeaheadData(\@dataToValidate);
	
	if ($result eq 1)
	{
		print "\nValidation passed.\n";
	}
	else
	{
		die "\nERROR: Validation failed.\n";
	}
}

# Remove the typeahead master property from DB to a file
sub DoCommand_remove
{
	# Export the master property into the file before remove it
	DoCommand_export();
	
	eval
	{
		$adminSessionObj->UnsetMasterProperty($propertyName, $dbName, $siteName);
		print "\nRemoved successfully.\n";
	};
	die "\nERROR: $@\n" if $@; $@ = "";
}

# Validate and clean the typeahead property stored in data
sub ValidateAndCleanTypeaheadData
{
	my @arg = @{$_[0]};

	if ($sessionObj->IsTextSearchEnabled() == 0)
	{
		return ValidateAndCleanTypeaheadData_againstSchema(\@arg);
	}
	else
	{
		return ValidateAndCleanTypeaheadData_againstFTSDeployment(\@arg);
	}
}

# Validate and clean the typeahead property stored in data
sub ValidateAndCleanTypeaheadData_againstSchema
{
	my @data = @{$_[0]};
	
	# Remove the blank items in the array
	@data = grep(/[^\s*]/, @data);
	
	# Get the indexed record types by FTS
	my @entityDefNames = $sessionObj->GetEntityDefNames();
	
	foreach my $item (@data)
	{
		my $rawItem = $item;
		my $status;
		my $record_type;
		my @fields;
		
		($status, $record_type, @fields) = processTypeaheadCfgItem($rawItem);
		
		if ($status == 0)
		{
			print "\nERROR: Missing record name before \"=\":\n".$rawItem."\n";
			return (0, undef);
		}
		
		# Check the record type and its fields are valid.
		if ($record_type~~@entityDefNames)
		{
			# Get the indexed field names of the record type
			my $entDef = $sessionObj->GetEntityDef($record_type);
			my $fieldList = $entDef->GetFieldDefNames();
			
			my $newItem = ""; # Used to construct a new item
			my $fieldInError;
			
			# Check the fields of the record type are in the schema
			($status, $newItem, $fieldInError) = processTypeaheadFields(\@fields, \@$fieldList);
			if ($status == 0)
			{
				print "\nERROR: The field \"".$fieldInError."\" in record type \"".$record_type."\" is not in your schema.\n".$rawItem."\n";
				return (0, undef);
			}
			
			$newItem = insertUniqueKeyFields($newItem, join(",", @{$entDef->GetDisplayNameFields()}));
			
			$item = $record_type."=".$newItem."\n";
		}
		else
		{
			print "\nERROR: The record type \"".$record_type."\" is not in your schema.\n".$rawItem."\n";
			return (0, undef);
		}
	}
	
	return (1, @data);
}

# Validate and clean the typeahead property stored in data against FTS deployment
sub ValidateAndCleanTypeaheadData_againstFTSDeployment
{
	my @data = @{$_[0]};
	
	# Remove the blank items in the array
	@data = grep(/[^\s*]/, @data);
	
	# Get the indexed record types by FTS
	my $textSearchDef = $sessionObj->BuildTextSearch();
	my @indexedEntityDefNames = $textSearchDef->GetIndexedEntityDefNames();
	
	foreach my $item (@data)
	{
		my $rawItem = $item;
		my $status;
		my $record_type;
		my @fields;
		
		($status, $record_type, @fields) = processTypeaheadCfgItem($rawItem);
		
		if ($status == 0)
		{
			print "\nERROR: Missing record name before \"=\":\n".$rawItem."\n";
			return (0, undef);
		}
		
		# Check the record type and its fields are valid.
		if ($record_type~~@indexedEntityDefNames)
		{
			# Get the indexed field names of the record type
			my $indexedFiledNames = $textSearchDef->GetIndexedFieldDefNames($record_type);
			
			# Check the unique keys for the record type are indexed by FTS
			my $entDef = $sessionObj->GetEntityDef($record_type);
			my $uniqueKeyFields = $entDef->GetDisplayNameFields();
			
			foreach my $unKey (@$uniqueKeyFields)
			{
				# Check the unique key field is indexed by FTS
				if (!($unKey~~@$indexedFiledNames))
				{
					print "\nERROR: The unique key \"".$unKey."\" in typeahead configuration data is not indexed by Full Text Search.\n".$rawItem."\n";
					return (0, undef);
				}
			}
			
			my $newItem = ""; # Used to construct a new item
			my $fieldInError;
			
			# Check the fields of the record type are indexed by FTS
			($status, $newItem, $fieldInError) = processTypeaheadFields(\@fields, \@$indexedFiledNames);
			if ($status == 0)
			{
				print "\nERROR: The field \"".$fieldInError."\" in record type \"".$record_type."\" is not indexed by Full Text Search.\n".$rawItem."\n";
				return (0, undef);
			}
			
			$newItem = insertUniqueKeyFields($newItem, join(",", @{$entDef->GetDisplayNameFields()}));
			
			$item = $record_type."=".$newItem."\n";
		}
		else
		{
			print "\nERROR: The record type \"".$record_type."\" is not indexed by Full Text Search.\n".$rawItem."\n";
			return (0, undef);
		}
	}
	
	return (1, @data);
}

# Given a typeahead data item, such as "Key=val1,Val2,Val#" validate and parse the data and 
# break it down into a set of "Key" and a list of "Val1,Val2,Val#".  In doing so, the funciton
# will also clean the data by normalizing it.
sub processTypeaheadCfgItem
{
	my $item = $_[0];
	
	my $rawItem = $item;
	my $newItem = ""; # Used to construct a new item
	
	$item =~ s/\s*//g;
	
	# Check the number of "=" in each line
	my $eqCharacter = "=";
	my $count = @{[$item =~ /$eqCharacter/g]}; ;  
	
	if ($count ne 1)
	{
		print "\nERROR: Only one \"=\" is allowed per line.\n".$rawItem."\n";
		return (0, undef, undef);
	}
	
	my @line = split("=", $item);
	
	# Check the name and fields of the record type exist
	if (!$line[0])
	{
		print "\nERROR: Missing record name before \"=\":\n".$rawItem."\n";
		return (0, undef, undef);
	}
	
	my $record_type = $line[0];
	
	my @fields;
	if ($line[1])
	{
		@fields = split(",", $line[1]);
	}
	
	return (1, $record_type, @fields);
}

# Given a list of fields, checks to see if they exist in a second set.  If no error
# is found, the returned string is a comman deliminated list of the fields.
sub processTypeaheadFields
{
	my ($fields, $fieldList) = @_;
	
	my $newItem = "";
	
	foreach (@$fields)
	{
		my $field = $_;
		
		if ($field ne "")
		{
			if (!($field~~@$fieldList))
			{
				# print "\nERROR: The field \"".$field."\" in record type \"".$record_type."\" is not in your schema.\n".$rawItem."\n";
				return (0, undef, $field);
			}
			else
			{
				# If the field is valid, add it into the newitem
				if ($newItem ne "")
				{
					$newItem .= ",";
				}
				$newItem .= $field;	# The first field don't need "," before it
			}
		}
	}
	
	return (1, $newItem, undef);
}

# Checks to see if the list of fields starts with the unique field(s), if not then the
# unique field(s) are added.
sub insertUniqueKeyFields
{
	my ($fields, $uniqueFields) = @_;
	
	if ($fields eq "")
	{
		# Case where there is no field list at all
		return $uniqueFields;
	}
	elsif ($fields eq $uniqueFields)
	{
		# Case where both field and unique field list are the same
		return $uniqueFields;
	}
	
	# Default case, field has data, but does it match the unique field?  If not, then insert it
	$uniqueFields .= ",";
	if (!($fields =~ /^$uniqueFields/))
	{
		$fields = $uniqueFields.$fields;
	}
	
	return $fields;
}

# Validate the user information and logon before process command line
sub ValidateAndLoginUser
{
	# Check admin session
 	eval
	{
		$adminSessionObj = CQAdminSession::Build();
		$adminSessionObj->Logon($user, $pwd, $dbSet);
	};
	die "\nERROR: $@\n" if $@; $@ = "";
	
	# Check user session
	eval
	{
		$sessionObj = CQSession::Build();
		$sessionObj->UserLogon($user, $pwd, $dbName, $dbSet);
	};
	die "\nERROR: $@\n" if $@; $@ = "";
	
	# Check super user
	if ($sessionObj->IsUserSuperUser() eq 0)
	{
		die "\nERROR: User has insufficient privileges.\n";
 	}
	
	# Check if FTS is enabled
	my $isFTSEnabled = $sessionObj->IsTextSearchEnabled();
	
	if (!$isFTSEnabled)
	{
		print "\nFull Text Search is not enabled; using schema data for validation.\n";
	}
	else
	{
		print "\nFull Text Search is enabled; using Full Text Search data for validation.\n";
	}
}
